package tsp;

import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.graphics.*;
import java.util.*;

public class City {
	int x;
	int y;
	TSPInstance tsp;
	String label;
	
	public City(TSPInstance tsp, int x, int y) {
		this.x = x;
		this.y = y;
		this.tsp = tsp;
	}
	
	public City(int x, int y) {
		this.x = x;
		this.y = y;
		tsp = null;
	}
	
	public void move(int x, int y) {
		this.x = x;
		this.y = y;
	}
	
	public City getClosestCity() {
		if (tsp == null) {
			return null;
		}
		
		double minDist = Double.MAX_VALUE;
		City closestCity = null;
		for (Enumeration e = tsp.cities.elements(); e.hasMoreElements();) {
			City c = (City) e.nextElement();
			double dist = distanceTo(c);
			if (dist > 0.0 &&  dist < minDist) {
				closestCity = c;
				minDist = dist;				
			}
		}
		return closestCity;
	}
	
	public double distanceTo(City c) {
		return Math.sqrt((x-c.x)*(x-c.x) + (y-c.y)*(y-c.y));	
	}
	
	public String toString() {
		return x+","+y;
	}
	
	public boolean equals(Object o) {
		return ((City)o).x==x && ((City)o).y==y;
	}
}
