package tsp;

import java.io.*;
import java.util.*;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Device;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.FormLayout;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.custom.CLabel;
import org.eclipse.swt.widgets.List;

public class MainWindow {

	private Shell sShell = null;
	private Menu menuBar = null;
	private Menu menuFile = null;
	private Menu menuTour = null;
	private Menu menuCities = null;
	private Menu menuAlgorithm = null;
	
	private MouseListener mouseListener = null;
	private MouseMoveListener mouseMoveListener = null;
	private City selectedCity = null;
	private MSTPyramid mst;
	private TSPInstance tsp;
	private Tour tour = null;
	private boolean isModified = false;
	private String filename = null;
	private final int DISPLAY_CITIES = 0;
	private final int INPUT_TOUR = 1;
	private final int DISPLAY_TOUR = 2;
	private int state = DISPLAY_CITIES;
	
	private boolean displayMeta = false;
	private boolean displayMetaChildren = false;
	
	private int maxClusterSize = 7;
	
	private TabFolder tabFolder = null;
	private List cityList = null;
	private CLabel statusBar = null;
	private CLabel cityLabel = null;
	private CLabel tourLabel = null;
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		Display display = Display.getDefault();
		MainWindow thisClass = new MainWindow();
		thisClass.createSShell();

		int border_width = thisClass.sShell.getSize().x - thisClass.sShell.getClientArea().width;
		int border_height = thisClass.sShell.getSize().y - thisClass.sShell.getClientArea().height - 42;
		
		thisClass.tsp = new TSPInstance();
		
		thisClass.sShell.setSize(512 + border_width + 20, 512 + 42+border_height +thisClass.tabFolder.getSize().y+thisClass.statusBar.getSize().y);
		thisClass.statusBar.setLocation(border_width/2, 512+border_height + thisClass.tabFolder.getSize().y);
		thisClass.tourLabel.setLocation(border_width/2+480, 512+border_height + thisClass.tabFolder.getSize().y);
		thisClass.cityLabel.setLocation(border_width/2+430, 512+border_height + thisClass.tabFolder.getSize().y);
		thisClass.tabFolder.setLocation(border_width/2, 512+border_height);
		thisClass.sShell.open();

		while (!thisClass.sShell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}
		display.dispose();
	}

	/**
	 * This method initializes the GUI
	 */
	private void createSShell() {
		sShell = new Shell(SWT.SHELL_TRIM & (~SWT.RESIZE));
		sShell.setText("MST-TSP Demo");
		sShell.setLayout(null);
		sShell.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		sShell.setSize(393, 286);
		statusBar = new CLabel(sShell,  SWT.LEFT | SWT.SHADOW_IN | SWT.SHADOW_OUT);
		statusBar.setBounds(0, 0, 430, 35);
		statusBar.setText("ready");
		statusBar.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		//statusBar.setFont(new Font(null,"Lucida",8,SWT.BOLD | SWT.ITALIC));
		
		cityLabel = new CLabel(sShell,  SWT.LEFT | SWT.SHADOW_IN | SWT.SHADOW_OUT);
		cityLabel.setBounds(0, 0, 50, 35);
		cityLabel.setText("0");
		cityLabel.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		cityLabel.setToolTipText("Current number of cities");

		tourLabel = new CLabel(sShell,  SWT.LEFT | SWT.SHADOW_IN | SWT.SHADOW_OUT);
		tourLabel.setBounds(0, 0, 50, 35);
		tourLabel.setText("0");
		tourLabel.setBackground(Display.getCurrent().getSystemColor(SWT.COLOR_WHITE));
		tourLabel.setToolTipText("Approximate length of current tour");
		
		tabFolder = new TabFolder(sShell, SWT.BORDER);
		tabFolder.setSize(530, 200);
		
		TabItem tab1 = new TabItem(tabFolder, 0);
		TabItem tab2 = new TabItem(tabFolder, 0);
		tab1.setText("Cities");
		tab2.setText("Pyramid");
		cityList = new List(tabFolder, SWT.V_SCROLL);
		tab1.setControl(cityList);
		Group optionsGroup = new Group(tabFolder, 0);
		optionsGroup.setLayout(new GridLayout(2, true));
		CLabel l1 = new CLabel(optionsGroup,0);
		l1.setText("Display meta-cities");
		Combo displayMetaCombo = new Combo(optionsGroup, 0);
		displayMetaCombo.add("no");
		displayMetaCombo.add("yes");
		displayMetaCombo.select(0);
		displayMetaCombo.addSelectionListener(new SelectionListener() {
			public void widgetDefaultSelected(SelectionEvent e) {}
			public void widgetSelected(SelectionEvent e) {
				displayMeta = (((Combo)e.widget).getSelectionIndex()==1);;
				citiesChanged();
			}
		});
		CLabel l2 = new CLabel(optionsGroup,0);
		l2.setText("Display meta-cities children");;
		Combo displayMetaChildrenCombo = new Combo(optionsGroup, 0);
		displayMetaChildrenCombo.add("no");
		displayMetaChildrenCombo.add("yes");
		displayMetaChildrenCombo.select(0);
		displayMetaChildrenCombo.addSelectionListener(new SelectionListener() {
			public void widgetDefaultSelected(SelectionEvent e) {}
			public void widgetSelected(SelectionEvent e) {
				displayMetaChildren = (((Combo)e.widget).getSelectionIndex()==1);
				citiesChanged();
			}
		});
		CLabel l3 = new CLabel(optionsGroup,0);
		l3.setText("Maximum cluster size");;
		Slider maxClusterSizeSlider = new Slider(optionsGroup, 0);
		maxClusterSizeSlider.setValues(7,2,10,1,1,1);
		maxClusterSizeSlider.addSelectionListener(new SelectionListener() {
			public void widgetDefaultSelected(SelectionEvent e) {}
			public void widgetSelected(SelectionEvent e) {
				if (mst != null) {
					maxClusterSize = ((Slider)e.widget).getSelection();
					citiesChanged();
				}
			}
		});

		tab2.setControl(optionsGroup);
		
		sShell.addPaintListener(new PaintListener() {
			public void paintControl(PaintEvent event) {
				if ((state == DISPLAY_TOUR || state == INPUT_TOUR) && tour != null && !tour.cities.isEmpty()) {
					event.gc.setLineWidth(2);
					for (int i = 0; i<tour.cities.size()-1; i++) {
						City c1 = tour.cities.get(i);
						City c2 = tour.cities.get(i+1);
						event.gc.drawLine(c1.x, c1.y, c2.x, c2.y);
					}
					if (state == DISPLAY_TOUR) {
						// close the tour
						City c1 = tour.cities.lastElement();
						City c2 = tour.cities.firstElement();
						event.gc.drawLine(c1.x, c1.y, c2.x, c2.y);
					}
				}
				
				// draw the cities of the TSP
				event.gc.setBackground(Display.getDefault().getSystemColor(SWT.COLOR_RED));
				event.gc.setForeground(Display.getDefault().getSystemColor(SWT.COLOR_BLACK));
				event.gc.setLineWidth(1);
				for (Iterator i = tsp.cities.iterator(); i.hasNext();) {
					City c = (City)i.next();
					
					event.gc.fillOval(c.x-10, c.y-10, 20, 20);
					event.gc.drawOval(c.x-10, c.y-10, 20, 20);
				}
				if (state == INPUT_TOUR && tour != null && !tour.cities.isEmpty()) {
					City c = tour.cities.lastElement();
					event.gc.setBackground(Display.getDefault().getSystemColor(SWT.COLOR_GREEN));
					event.gc.setForeground(Display.getDefault().getSystemColor(SWT.COLOR_BLACK));
					event.gc.setLineWidth(2);
					event.gc.fillOval(c.x-10, c.y-10, 20, 20);
					event.gc.drawOval(c.x-10, c.y-10, 20, 20);
				}
				
				// TODO experimental, only for MST pyramid
				if (displayMeta && mst != null) {
					event.gc.setAlpha(110);
					
					for (int j = 1; j < mst.tsps.size(); j++) {
						for (Iterator i = mst.tsps.get(j).cities.iterator(); i.hasNext();) {
							City c = (City)i.next();
							
							if (j==1)
								event.gc.setBackground(Display.getDefault().getSystemColor(SWT.COLOR_YELLOW));
							else if (j==2)
								event.gc.setBackground(Display.getDefault().getSystemColor(SWT.COLOR_GREEN));
							else
								event.gc.setBackground(Display.getDefault().getSystemColor(SWT.COLOR_BLUE));
							event.gc.setLineWidth(1);
							event.gc.fillOval(c.x-10, c.y-10, 20, 20);
							event.gc.drawOval(c.x-10, c.y-10, 20, 20);
							
							if (displayMetaChildren) {
								event.gc.setLineWidth(j*2-1);
								// draw lines to children
								for (Iterator k = mst.children.get(c).iterator(); k.hasNext();) {
									City child = (City)k.next();
									event.gc.drawLine(c.x, c.y, child.x, child.y);
								}
							}
						}
					}
				}
			}
		});
		
		menuBar = new Menu(sShell, SWT.BAR);
		MenuItem menuItemFile = new MenuItem(menuBar, SWT.CASCADE);
		MenuItem menuItemCities = new MenuItem(menuBar, SWT.CASCADE);
		MenuItem menuItemTour = new MenuItem(menuBar, SWT.CASCADE);
		MenuItem menuItemAlgorithm = new MenuItem(menuBar, SWT.CASCADE);
		menuItemFile.setText("File");
		menuItemCities.setText("Cities");
		menuItemTour.setText("Tour");
		menuItemAlgorithm.setText("Algorithm");
		
		menuFile = new Menu(menuItemFile);
		menuCities = new Menu(menuItemCities);
		menuTour = new Menu(menuItemTour);
		menuAlgorithm = new Menu(menuItemAlgorithm);
		
		menuItemFile.setMenu(menuFile);
		menuItemCities.setMenu(menuCities);
		menuItemTour.setMenu(menuTour);
		menuItemAlgorithm.setMenu(menuAlgorithm);
		

		// File menu
		MenuItem pushFileNew = new MenuItem(menuFile, SWT.PUSH);
		pushFileNew.setText("New");
		pushFileNew.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				newTSP();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		MenuItem pushFileOpen = new MenuItem(menuFile, SWT.PUSH);
		pushFileOpen.setText("Open");
		pushFileOpen.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				loadTSP();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		MenuItem pushFileSave = new MenuItem(menuFile, SWT.PUSH);
		pushFileSave.setText("Save");
		pushFileSave.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				if (filename == null)
					saveAsTSP();
				else
					saveTSP();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		MenuItem pushFileSaveAs = new MenuItem(menuFile, SWT.PUSH);
		pushFileSaveAs.setText("Save As");
		pushFileSaveAs.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				saveAsTSP();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		MenuItem separator = new MenuItem(menuFile, SWT.SEPARATOR);
		MenuItem pushFileExit = new MenuItem(menuFile, SWT.PUSH);
		pushFileExit.setText("Exit");
		pushFileExit.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				sShell.dispose();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		

		// Cities menu
		MenuItem pushMenuAddCity = new MenuItem(menuCities, SWT.PUSH);
		pushMenuAddCity.setText("Add City");
		pushMenuAddCity.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				addCity();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		MenuItem pushMenuRemoveCity = new MenuItem(menuCities, SWT.PUSH);
		pushMenuRemoveCity.setText("Remove City");
		pushMenuRemoveCity.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				removeCity();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		MenuItem pushMenuMoveCity = new MenuItem(menuCities, SWT.PUSH);
		pushMenuMoveCity.setText("Move City");
		pushMenuMoveCity.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				moveCity();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});

		
		// Tour menu
		MenuItem pushTourLoad = new MenuItem(menuTour, SWT.PUSH);
		pushTourLoad.setText("Load from File");
		pushTourLoad.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				loadTour();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		MenuItem pushTourSave = new MenuItem(menuTour, SWT.PUSH);
		pushTourSave.setText("Save to File");
		pushTourSave.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				saveAsTour();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});
		MenuItem pushTourInput = new MenuItem(menuTour, SWT.PUSH);
		pushTourInput.setText("Input Tour");
		pushTourInput.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				inputTour();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});

		
		// Algorithm menu
		MenuItem pushAlgorithmMST = new MenuItem(menuAlgorithm, SWT.PUSH);
		pushAlgorithmMST.setText("MST-Pyramid");
		pushAlgorithmMST.addSelectionListener(new org.eclipse.swt.events.SelectionListener() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				state = DISPLAY_TOUR;
				tourChanged();
			}
			public void widgetDefaultSelected(org.eclipse.swt.events.SelectionEvent e) {
			}
		});

		sShell.setMenuBar(menuBar);
	}
	
	public void loadTSP() {
		// if the current TSP was modified -> ask to save it before throwing it away
		if (isModified) {
			MessageBox mb = new MessageBox(sShell, SWT.YES | SWT.NO | SWT.CANCEL);
			mb.setMessage("The current TSP was modified - do you want to save the changes?");
			switch (mb.open()) {
			case SWT.CANCEL: return;
			case SWT.YES: saveTSP();
			}
		}

		FileDialog fd = new FileDialog(sShell, SWT.OPEN);
		fd.setFilterExtensions(new String[] {"*.tsp"});
		fd.setFilterPath("/data/tsp-series");
		filename = fd.open();
		if (filename != null) { 
			try {
				tsp = new TSPInstance(filename);
				citiesChanged();
			} catch(IOException e) {
				MessageBox mb = new MessageBox(sShell, SWT.ICON_ERROR);
				mb.setMessage("Error opening file!");
				mb.open();
			}
		}
	}

	public boolean saveTSP() {
		if (filename != null) {
			try {
				tsp.saveToFile(filename);
			} catch(IOException e) {
				MessageBox mb = new MessageBox(sShell, SWT.ICON_ERROR);
				mb.setMessage("Error saving file!");
				mb.open();
				return false;
			}
			return true;
		}
		return false;
	}
	
	public void saveAsTSP() {
		FileDialog fd = new FileDialog(sShell, SWT.SAVE);
		fd.setFilterExtensions(new String[] {"*.tsp"});
		fd.setFilterPath("/data/tsp-series");
		fd.setFileName("unnamed.tsp");
		filename = fd.open();

		// check for existing file
		if (filename != null && new File(filename).exists()) {
			MessageBox mb = new MessageBox(sShell, SWT.YES | SWT.NO);
			mb.setMessage("The file already exists - do you want to overwrite?");
			if (mb.open() == SWT.NO)
				filename = null;
		}
		saveTSP();
	}

	public void saveAsTour() {
		FileDialog fd = new FileDialog(sShell, SWT.SAVE);
		fd.setFilterExtensions(new String[] {"*.tour"});
		fd.setFilterPath("/data/tsp-series");
		fd.setFileName("unnamed.tour");
		String tourFilename = fd.open();

		// check for existing file
		if (tourFilename != null && new File(tourFilename).exists()) {
			MessageBox mb = new MessageBox(sShell, SWT.YES | SWT.NO);
			mb.setMessage("The file already exists - do you want to overwrite?");
			if (mb.open() == SWT.NO)
				return;
		}
		if (tourFilename != null) {
			try {
				tour.saveToFile(tourFilename);
			} catch(IOException e) {
				MessageBox mb = new MessageBox(sShell, SWT.ICON_ERROR);
				mb.setMessage("Error saving file!");
				mb.open();
			}
		}
	}

	public void loadTour() {
		FileDialog fd = new FileDialog(sShell, SWT.OPEN);
		fd.setFilterExtensions(new String[] {"*.tour"});
		fd.setFilterPath("/data/tsp-series");
		String tourFilename = fd.open();
		if (tourFilename != null) { 
			try {
				tour = new Tour(tsp, tourFilename);
				state = DISPLAY_TOUR;
				tourChanged();
			} catch(IOException e) {
				MessageBox mb = new MessageBox(sShell, SWT.ICON_ERROR);
				mb.setMessage("Error opening file!");
				mb.open();
			}
		}
	}
	
	public void newTSP() {
		// if the current TSP was modified -> ask to save it before throwing it away
		if (isModified) {
			MessageBox mb = new MessageBox(sShell, SWT.YES | SWT.NO | SWT.CANCEL);
			mb.setMessage("The current TSP was modified - do you want to save the changes?");
			switch (mb.open()) {
			case SWT.CANCEL: return;
			case SWT.YES: saveTSP();
			}
		}
		tsp = new TSPInstance();
		tour = null;
		state = DISPLAY_CITIES;
		citiesChanged();
	}
	
	public void addCity() {
		statusBar.setText("Click to add cities");
		
		if (mouseListener != null) {
			sShell.removeMouseListener(mouseListener);
		}
		mouseListener = new org.eclipse.swt.events.MouseAdapter() {
			public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
				tsp.cities.add(new City(tsp, e.x, e.y));
				citiesChanged();
			}
		}; 
		sShell.addMouseListener(mouseListener);
	}
	
	public void removeCity() {
		statusBar.setText("Click to select city to remove");
		
		if (mouseListener != null) {
			sShell.removeMouseListener(mouseListener);
		}

		mouseListener = new org.eclipse.swt.events.MouseAdapter() {
			public void mouseDown(org.eclipse.swt.events.MouseEvent e) {
				City c = coordinatesToCity(e.x, e.y);
				
				if (c != null) {
					tsp.cities.remove(c);
					citiesChanged();
				}
			}
		};
		sShell.addMouseListener(mouseListener);
	}

	public void moveCity() {
		statusBar.setText("Click to select city and drag it to a new position");
		
		if (mouseListener != null) {
			sShell.removeMouseListener(mouseListener);
		}
		if (mouseMoveListener != null) {
			sShell.removeMouseMoveListener(mouseMoveListener);
		}

		mouseListener = new org.eclipse.swt.events.MouseAdapter() {
			public void mouseDown(org.eclipse.swt.events.MouseEvent e) {
				 selectedCity = coordinatesToCity(e.x, e.y);
			}
			
			public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
				if (selectedCity != null) {
					selectedCity.move(e.x, e.y);
					selectedCity = null;
					citiesChanged();
				}
			}
		};
		mouseMoveListener = new org.eclipse.swt.events.MouseMoveListener() {
			public void mouseMove(org.eclipse.swt.events.MouseEvent e) {
				if (selectedCity != null) {
					selectedCity.move(e.x, e.y);
					citiesChanged();
				}
			}
		};
		sShell.addMouseListener(mouseListener);
		sShell.addMouseMoveListener(mouseMoveListener);
	}

	public void inputTour() {
		tour = new Tour(tsp);
		state = INPUT_TOUR;
		statusBar.setText("Click on the cities to add them to the tour");
		sShell.redraw();
		
		if (mouseListener != null) {
			sShell.removeMouseListener(mouseListener);
		}
		mouseListener = new org.eclipse.swt.events.MouseAdapter() {
			public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
				City c = coordinatesToCity(e.x, e.y);
				
				if (c != null) {
					 try {
						 tour.append(c);
					 } catch (TSPException ex) {
					 }
					 if (tour.cities.size() == tsp.cities.size()) {
						 state = DISPLAY_TOUR;
						 statusBar.setText("ready");
					 }
				}
				tourLabel.setText(""+(int)tour.getLength());
				sShell.redraw();
			}
		};
		sShell.addMouseListener(mouseListener);
	}
	
	public City coordinatesToCity(int x, int y) {
		// TODO: this could probably be done more efficiently
		int cx,cy,i;
		for (i=0; i<tsp.cities.size(); i++) {
			cx = tsp.cities.get(i).x;
			cy = tsp.cities.get(i).y;
			if (Math.abs(cx - x)<10 && Math.abs(cy - y)<10) {
				return tsp.cities.get(i);
			}
		}
		return null;
	}
	
	public void citiesChanged() {
		isModified = true;
		cityList.removeAll();
		for (int i=0; i<tsp.cities.size(); i++) {
			cityList.add(tsp.cities.get(i).toString());
		}
		cityLabel.setText(""+tsp.cities.size());
		tourChanged();
	}
	
	public void tourChanged() {
		if (state == DISPLAY_TOUR) {
			mst = new MSTPyramid(tsp);
			mst.maxClusterSize = maxClusterSize; 
			mst.cluster();
			mst.refine();
			if (mst.tour != null) {
				tour = mst.tour;
				tourLabel.setText(""+(int)tour.getLength());
			}
		}
		sShell.redraw();
	}
}
