package tsp;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.reflect.Array;
import java.util.*;
import java.io.*;

/**
 * This class represents a symmetric instance of the travelling salesman problem
 * @author Andreas Lehrbaum
 */
public class TSPInstance {
	Vector<City> cities;
	double[][] distanceMatrix;
	int width;
	int height;
	
	public TSPInstance() {
		cities = new Vector<City>();
	}
	
	public TSPInstance(String filename) throws IOException {
		cities = new Vector<City>();
		
		DataInputStreamLE dis = null;
		dis = new DataInputStreamLE(new FileInputStream(filename));
	
		// for legacy reasons
		int version=0, num_of_cities=0;

		version = dis.readIntLE();
		width = dis.readIntLE();
		height = dis.readIntLE();
		num_of_cities = dis.readIntLE();
		
		// read the coordinates of the cities and add them to the TSPInstance
		while (dis.available() >= 8) {
			addCity(new City(dis.readIntLE(), dis.readIntLE()));
		}
		dis.close();
	}
	
	public void saveToFile(String filename) throws IOException {
		DataOutputStreamLE dos = null;
		System.out.println("Trying to read from "+filename);
		dos = new DataOutputStreamLE(new FileOutputStream(filename));
	
		int version=1;

		dos.writeIntLE(version);
		dos.writeIntLE(width);
		dos.writeIntLE(height);
		dos.writeIntLE(cities.size());
		
		// write the coordinates of the cities to the file
		for (Iterator i=cities.iterator();i.hasNext();) {
			City c = (City)i.next();
			dos.writeIntLE(c.x);
			dos.writeIntLE(c.y);
		}
		// for compatibility reasons: store a zero meaning no obstacles
		dos.writeIntLE(0);
		dos.close();
	}
	
	public void addCity(City c) {
		cities.add(c);
		c.tsp = this;
	}
	
	public void updateDistanceMatrix() {
		// check if the number of cities has changed -> if so, we need to allocate a new distanceMatrix
		if (distanceMatrix == null || distanceMatrix.length != cities.size()) {
			distanceMatrix = new double[cities.size()][cities.size()];
		}
		// fill the distanceMatrix with the distances
		// NOTE: we rely on symmetry here
		for (int i = 0; i < cities.size()-1; i++) {
			for (int j = i+1; j < cities.size(); j++) {
				System.out.println("i: "+i+" j: "+j);
				distanceMatrix[i][j] = cities.get(i).distanceTo(cities.get(j));
				distanceMatrix[j][i] = distanceMatrix[i][j];
			}
		}
	}
	
	public double getDistance(int i, int j) {
		return distanceMatrix[i][j];
	}
}
