package tsp;

import java.util.Vector;
import java.io.*;
import java.util.Iterator;


public class Tour {
	TSPInstance tsp;
	public Vector <City> cities;
	
	public Tour(TSPInstance tsp) {
		this.tsp = tsp;
		cities = new Vector<City>();
	}
	
	public Tour(TSPInstance tsp, String filename) throws IOException {
		this.tsp = tsp;
		cities = new Vector<City>();
		
		DataInputStreamLE dis = null;
		dis = new DataInputStreamLE(new FileInputStream(filename));
	
		int size = dis.readIntLE();
		
		// read the coordinates of the cities from the file
		for (int i=0;i<size;i++) {
			int x = dis.readIntLE();
			int y = dis.readIntLE();
			City c = new City(x,y);
			// if we encounter a city non-existant in our TSPInstance -> abort
			if (!tsp.cities.contains(c)) {
				cities.clear();
				throw new IOException("Tour not suitable for this TSP!");
			}
			cities.add(c);
		}
		dis.close();
	}
	
	// add a city at the end of the tour
	public void append(City c) throws TSPException {
		if (tsp.cities.contains(c)) {
			if (!cities.contains(c)) {
				cities.add(c);
			} else {
				throw new TSPException("City allready in tour!");
			}
		} else {
			throw new TSPException("City not in TSP");
		}
	}
	
	public void insertAt(City c, int i) throws TSPException {
		if (tsp.cities.contains(c)) {
			cities.insertElementAt(c, i);
		} else {
			throw new TSPException("City not in TSP");
		}
	}
	
	public void insertAfter(City c, City reference) throws TSPException {
		if (tsp.cities.contains(c) && tsp.cities.contains(reference)) {
			cities.insertElementAt(c,cities.indexOf(reference)+1);
		} else {
			throw new TSPException("City not in TSP");
		}
	}
	
	public void insertBefore(City c, City reference) throws TSPException {
		if (tsp.cities.contains(c) && tsp.cities.contains(reference)) {
			cities.insertElementAt(c,cities.indexOf(reference));
		} else {
			throw new TSPException("City not in TSP");
		}
	}
	
	public double getLength() {
		double length = 0.0;
		if (cities.size() > 1) {
			for (int i = 0; i<cities.size()-1; i++) {
				City c1 = cities.get(i);
				City c2 = cities.get(i+1);
				length += c1.distanceTo(c2);
			}
			// if our tour does not end where it began -> add the remaining distance 
			if (! cities.firstElement().equals(cities.lastElement())) {
				City c1 = cities.firstElement();
				City c2 = cities.lastElement();
				length += c1.distanceTo(c2);
			}
		}
		return length;
	}
	
	public void saveToFile(String filename) throws IOException {
		DataOutputStreamLE dos = null;
		dos = new DataOutputStreamLE(new FileOutputStream(filename));
	
		dos.writeIntLE(cities.size());
		
		// write the coordinates of the cities to the file
		for (Iterator i=cities.iterator();i.hasNext();) {
			City c = (City)i.next();
			dos.writeIntLE(c.x);
			dos.writeIntLE(c.y);
		}
		dos.close();
	}
	
	public String toString() {
		StringBuffer buf = new StringBuffer(cities.size()+" cities, length="+Math.round(getLength())+": ");
		for (Iterator i=cities.iterator();i.hasNext();) {
			buf.append(i.next().toString()+" ");
		}
		return buf.toString();
	}
}
